package net.minestom.server.crypto;

import net.minestom.server.utils.binary.BinaryReader;
import net.minestom.server.utils.binary.BinaryWriter;
import net.minestom.server.utils.binary.Writeable;
import net.minestom.server.utils.crypto.KeyUtils;

import java.nio.charset.StandardCharsets;
import java.security.PublicKey;
import java.time.Instant;
import java.util.Arrays;

/**
 * Player's public key used to sign chat messages
 *
 * @param expiresAt signature expiration date
 * @param publicKey the player's public key
 * @param signature {@link #signedPayload()} signed by Yggdrasil
 */
public record PlayerPublicKey(Instant expiresAt, PublicKey publicKey, byte[] signature) implements Writeable {

    public PlayerPublicKey(BinaryReader reader) {
        this(Instant.ofEpochMilli(reader.readLong()),
                KeyUtils.publicRSAKeyFrom(reader.readByteArray()), reader.readByteArray());
    }

    public byte[] signedPayload() {
        return (expiresAt.toEpochMilli() + KeyUtils.rsaPublicKeyToString(publicKey)).getBytes(StandardCharsets.US_ASCII);
    }

    public boolean isValid() {
        return expiresAt.isAfter(Instant.now()) && SignatureValidator.YGGDRASIL.validate(signedPayload(), signature());
    }

    @Override
    public void write(BinaryWriter writer) {
        writer.writeLong(expiresAt().toEpochMilli());
        writer.writeByteArray(publicKey.getEncoded());
        writer.writeByteArray(signature());
    }

    @Override
    public boolean equals(Object obj) {
        if (obj instanceof PlayerPublicKey ppk) {
            return expiresAt.equals(ppk.expiresAt) && publicKey.equals(ppk.publicKey) && Arrays.equals(signature, ppk.signature);
        } else {
            return false;
        }
    }
}
